//-------------------------------------------------- Ver1.0.0 ------ 2021-01-07 --------------
// Acoustic Target Systems
//----------------------------------------- written by Hiroyuki Iizuka ---- Japan ------------

#define DEBUG

#include <TFT_HX8357_Due.h>              // Hardware-specific library
TFT_HX8357_Due tft = TFT_HX8357_Due();   // Invoke custom library
#define TFT_GREY 0x5AEB

#define TIMER_FREQENCY  16000000         // PIC24FJ64G002 32MHz fosc/2

#define COMMAND_QUERY_TEMPERATURE     '1'
#define COMMAND_LED1_ON               '2'
#define COMMAND_LED1_OFF              '3'
#define COMMAND_RELAY1_ON             '4'
#define COMMAND_RELAY1_OFF            '5'
#define COMMAND_RELAY2_ON             '6'
#define COMMAND_RELAY2_OFF            '7'

#define pow2(a) ((a)*(a))

#define RS485_CONTROL_PIN  2
#define PUSH_SW1           3
#define PUSH_SW2           4
#define PUSH_SW3           5
#define LED1               6
#define LED2               7 
#define DIP_SW1            8
#define DIP_SW2            9
#define DIP_SW3           10
#define DIP_SW4           11

#define SW_ON     LOW
#define SW_OFF    HIGH
#define LED_ON    HIGH
#define LED_OFF   LOW

struct mic_coord {
 double  x;
 double  y;
 double  r;
};

struct point_of_impact {
 double  x;
 double  y;
};

//--------- USER EDIT BEGIN --------------------------------------------------
#define NUMBER_OF_MIC   4      // 3 or 4

#define CALIBER                   6.0   // mm
#define TARGET_DIAMETER_10POINT  11.5   // mm
#define TARGET_DIAMETER_STEP     16.0   // mm
#define TARGET_AREA_BLACK           7   // point
#define TARGET_DOT_SIZE          0.25   // 1dot = 0.25mm  

mic_coord mic[] = {
  {  -0.71,  -0.71 },    // mic No1 X coordinate, mic No1 Y coordinate
  { 336.71,  -0.71 },    // mic No2 X coordinate, mic No2 Y coordinate
  {  -0.71, 332.71 },    // mic No3 X coordinate, mic No3 Y coordinate
  { 336.71, 332.71 }     // mic No4 X coordinate, mic No4 Y coordinate
};

#define TARGET_Z           11.0   // mm 
#define TARGET_CENTER_X   168.0   // mm
#define TARGET_CENTER_Y   166.0   // mm
#define TARGET_WIDTH      260.0   // mm
#define TARGET_HIGHT      160.0   // mm
//--------- USER EDIT END ----------------------------------------------------

boolean       Mic_error;
unsigned int  timer_count[NUMBER_OF_MIC];
String        serial_buf;
unsigned char record_no;
double        temperature;
boolean       first_bullet;
boolean       switch1_enable;
boolean       switch2_enable;
boolean       switch3_enable;

//-----------------------------------------------------------------------------------------
// draw air pistol target
//-----------------------------------------------------------------------------------------
void draw_AP_target()
{
  double target_radius;
  double text_radius;
  uint16_t circle_color;
  uint16_t center_x,center_y;
  
  center_x = HX8357_TFTHEIGHT / 2;
  center_y = HX8357_TFTWIDTH  / 2;

  tft.fillScreen(TFT_GREY);
  circle_color = TFT_BLACK;
  for( int i = 1 ; i <= 10 ; i++ ) { 
    target_radius =  (TARGET_DIAMETER_10POINT + TARGET_DIAMETER_STEP * (10 - i)) / 2 / TARGET_DOT_SIZE; 
    text_radius   =  (TARGET_DIAMETER_10POINT + TARGET_DIAMETER_STEP * (10 - i) - TARGET_DIAMETER_STEP / 2) / 2 / TARGET_DOT_SIZE; 
    if( i == TARGET_AREA_BLACK ) {
      circle_color = TFT_WHITE;
      tft.fillCircle(center_x, center_y, (uint16_t)target_radius , TFT_BLACK );
    }
    tft.drawCircle(center_x, center_y, (uint16_t)target_radius , circle_color );
    if( i <= 6 ) {
      tft.setTextColor(TFT_BLACK, TFT_GREY);      
    } else {
      tft.setTextColor(TFT_WHITE, TFT_BLACK);
    }
    if( 3 <= i && i <= 9 ) {
      tft.drawNumber( i, center_x + (uint16_t)text_radius -7, center_y -10 , 4);
      tft.drawNumber( i, center_x - (uint16_t)text_radius -7, center_y -10 , 4);
    }
  }
}

//-----------------------------------------------------------------------------------------
// led and switch routine
//-----------------------------------------------------------------------------------------
void led1_on()  { digitalWrite( LED1, LED_ON  ); }
void led1_off() { digitalWrite( LED1, LED_OFF ); }
void led2_on()  { digitalWrite( LED2, LED_ON  ); }
void led2_off() { digitalWrite( LED2, LED_OFF ); }

boolean sw1_on()
{
  if( switch1_enable ) {
    if( digitalRead( PUSH_SW1 ) == SW_ON ) {
      switch1_enable = false;
      return true;
    } else {
      return false;
    }  
  }
}

boolean sw2_on()
{
  if( switch2_enable ) {
    if( digitalRead( PUSH_SW2 ) == SW_ON ) {
      switch2_enable = false;
      return true;
    } else {
      return false;
    }  
  }
}

boolean sw3_on()
{
  if( switch3_enable ) {
    if( digitalRead( PUSH_SW3 ) == SW_ON ) {
      switch3_enable = false;
      return true;
    } else {
      return false;
    }  
  }
}

//-----------------------------------------------------------------------------------------
// draw bullet mark
// x:x coordinate[mm]
// y:y coordinate[mm]
// caliber:bullet diameter[mm]
// inner_color: TFT_RED or TFT_GREY
//-----------------------------------------------------------------------------------------
void draw_bullet(double x, double y, double caliber, uint16_t inner_color)
{
  uint16_t tft_x;
  uint16_t tft_y;
  uint16_t radius;
  
  radius = (uint16_t)(caliber / 2 / TARGET_DOT_SIZE);
  tft_x = HX8357_TFTHEIGHT / 2 + (int)(x / TARGET_DOT_SIZE);
  tft_y = HX8357_TFTWIDTH  / 2 - (int)(y / TARGET_DOT_SIZE);
  tft.drawCircle(tft_x, tft_y, radius,     TFT_WHITE );
  tft.fillCircle(tft_x, tft_y, radius - 1, inner_color );  
}
 
//-----------------------------------------------------------------------------------------
// debug print
//-----------------------------------------------------------------------------------------
void dump_circle( double x, double y, double r, unsigned char penno )
{
  char str[60];

  Serial.print( "PenNo(" );
  Serial.print( penno );
  Serial.print( "); " );
  Serial.print( "Circle(" );
  Serial.print( x );
  Serial.print( "," );
  Serial.print( y );
  Serial.print( "," );
  Serial.print( r );
  Serial.print( ");" );
}

void dump_double( double arg )
{
  char str[20];

  sprintf( str, "%10.4f", arg );
  Serial.print( str );
}

void dump_temperature( double arg )
{
  char str[10];

  sprintf( str, "%4.1f", arg );
  Serial.println( str );
}

void dump_mm( double arg )
{
  char str[10];

  sprintf( str, "%6.2f", arg );
  Serial.print( str );
}

void dump_uint( unsigned int arg )
{
  char buf[10];

  sprintf( buf, "%6d", arg );
  Serial.print( buf );
}


//-----------------------------------------------------------------------------------------
// calc temperature -> speed of sound
//-----------------------------------------------------------------------------------------
double speed_of_sound( double air_temperature )
{
  return (331.5 + 0.6 * air_temperature) * 1000;
}

//-----------------------------------------------------------------------------------------
// calc timer count -> distance mm
//-----------------------------------------------------------------------------------------
double calc_distance( unsigned int timer_count, double temperature )
{
  return (double)timer_count * speed_of_sound( temperature ) / TIMER_FREQENCY;
}

//-----------------------------------------------------------------------------------------
// test atrget area
//-----------------------------------------------------------------------------------------
boolean test_target_area( point_of_impact poi )
{
  double target_area_left;
  double target_area_right;
  double target_area_top;
  double target_area_bottom;

  target_area_left   = TARGET_CENTER_X - TARGET_WIDTH / 2;
  target_area_right  = TARGET_CENTER_X + TARGET_WIDTH / 2;
  target_area_top    = TARGET_CENTER_Y + TARGET_HIGHT / 2;
  target_area_bottom = TARGET_CENTER_Y - TARGET_HIGHT / 2;
 
  if( (target_area_left   <= poi.x) && (poi.x <= target_area_right) && 
      (target_area_bottom <= poi.y) && (poi.y <= target_area_top) ) {
    return true;
  } else {
    return false;
  }
}

//-----------------------------------------------------------------------------------------
// calc point of impact
//-----------------------------------------------------------------------------------------
struct point_of_impact calc_not_mic_arrange_linear( 
  struct mic_coord mic1,
  struct mic_coord mic2,
  struct mic_coord mic3,
  double target_z )
{
  double a1,a2,a3;
  double b1,b2,b3;
  double c1,c2,c3;
  double d1,d2,d3;
  double M,N,P,Q;
  double a,b,c;
  double sqr;
  double r1,r2;
  double xs1,xs2,ys1,ys2;
  point_of_impact result;
  
  a1 = 2 * ( mic2.x - mic1.x );
  a2 = 2 * ( mic3.x - mic2.x );
  a3 = 2 * ( mic1.x - mic3.x );
  b1 = 2 * ( mic2.y - mic1.y );
  b2 = 2 * ( mic3.y - mic2.y );
  b3 = 2 * ( mic1.y - mic3.y );
  c1 = 2 * ( mic2.r - mic1.r );
  c2 = 2 * ( mic3.r - mic2.r );
  c3 = 2 * ( mic1.r - mic3.r );
  d1 = pow2(mic1.r) - pow2(mic2.r) + pow2(mic2.x) - pow2(mic1.x) + pow2(mic2.y) - pow2(mic1.y);
  d2 = pow2(mic2.r) - pow2(mic3.r) + pow2(mic3.x) - pow2(mic2.x) + pow2(mic2.y) - pow2(mic2.y);
  d3 = pow2(mic3.r) - pow2(mic1.r) + pow2(mic1.x) - pow2(mic3.x) + pow2(mic1.y) - pow2(mic3.y);
  M = ( d2 * b1 - b2 * d1 ) / ( a2 * b1 - a1 * b2 );
  N = ( b2 * c1 - b1 * c2 ) / ( a2 * b1 - a1 * b2 );
  P = ( d3 * a1 - a3 * d1 ) / ( a1 * b3 - a3 * b1 );
  Q = ( a3 * c1 - a1 * c3 ) / ( a1 * b3 - a3 * b1 );
  a = pow2( N ) + pow2( Q ) - 1;
  b = 2 * ( M * N - N * mic1.x + P * Q - Q * mic1.y - mic1.r );
  c = pow2( M ) - 2 * M * mic1.x + pow2(mic1.x) + pow2( P ) - 2 * P * mic1.y + pow2(mic1.y) - pow2(mic1.r) + pow2( target_z ); 
  
  sqr = sqrt(pow2(b)-4*a*c);
  r1 = (-b+sqr)/(2*a);
  r2 = (-b-sqr)/(2*a);
  xs1 = M + N * r1;
  xs2 = M + N * r2;
  ys1 = P + Q * r1;
  ys2 = P + Q * r2;

  result.x = xs1;
  result.y = ys1;

  if( (r1 < 0) || (test_target_area( result ) == false) ) {
    result.x = xs2;
    result.y = ys2;
  }
  if( (r2 < 0) || (test_target_area( result ) == false) ) {
    result.x = -1;
    result.y = -1;
    led2_on();
    #ifdef DEBUG
    Serial.println("// out of target");
    #endif
  }
  return result;
}  

//-----------------------------------------------------------------------------------------
// calc point of impact mic arrange linerly 
//-----------------------------------------------------------------------------------------
struct point_of_impact calc_mic_arrange_linear_sub( 
  mic_coord mic1, 
  mic_coord mic2, 
  mic_coord mic3, 
  double target_z )
{
  double a,b,c,d;
  double tmp;
  double rs;
  double Xs,Ys;
  point_of_impact result;

  a = ( mic1.r - mic2.r ) / ( mic2.x - mic1.x );
  b = ( pow2(mic1.r) - pow2(mic2.r) + pow2(mic2.x) - pow2(mic1.x) ) / ( 2 * ( mic2.x - mic1.x ));
  c = ( mic2.r - mic3.r ) / ( mic3.x - mic2.x );
  d = ( pow2(mic2.r) - pow2(mic3.r) + pow2(mic3.x) - pow2(mic2.x) ) / ( 2 * ( mic3.x - mic2.x ));

  rs = ( d - b ) / ( a - c );
  result.x = c * rs + d;
  result.y = sqrt( pow2(rs + mic1.r) - pow2(result.x - mic1.x) - pow2(target_z) );

  return result;
}

//-----------------------------------------------------------------------------------------
// test mic arrange
//-----------------------------------------------------------------------------------------
boolean test_mic_arrange_liner(unsigned char id1, unsigned char id2, unsigned char id3 )
{
  double a1,a2,b1,b2;
    
  a1 = ( mic[id2].x - mic[id1].x );
  a2 = ( mic[id3].x - mic[id2].x );
  b1 = ( mic[id2].y - mic[id1].y );
  b2 = ( mic[id3].y - mic[id2].y );
  if( (a2 * b1 - a1 * b2) == 0 ) {    // linear
    return true;
  } else {
    return false;
  }
}

//-----------------------------------------------------------------------------------------
// calc point of impact mic arrange linerly
//-----------------------------------------------------------------------------------------
point_of_impact calc_mic_arrange_linear( 
  mic_coord mic1, 
  mic_coord mic2, 
  mic_coord mic3, 
  double target_z )
{
  double offset_x, offset_y;
  double rad;
  mic_coord mic_tmp1;
  mic_coord mic_tmp2;
  mic_coord mic_tmp3;
  point_of_impact poi, poi1, poi2;

  mic_tmp1.r = mic1.r;
  mic_tmp2.r = mic2.r;
  mic_tmp3.r = mic3.r;
  mic_tmp1.y = 0.0;
  mic_tmp2.y = 0.0;
  mic_tmp3.y = 0.0;
  mic_tmp1.x = 0.0;
  offset_x = mic1.x;
  offset_y = mic1.y;
  if( mic1.x == mic2.x ) {   // vertial
    mic_tmp2.x = mic2.y - offset_y;
    mic_tmp3.x = mic3.y - offset_y;
    rad = M_PI_2;   // M_PI_2 = pi / 2
  } else {
    mic_tmp2.x = sqrt( pow2(mic2.x - offset_x) + pow2(mic2.y - offset_y) );
    if( mic2.x < offset_x ) { 
      mic_tmp2.x = -mic_tmp2.x;
    }  
    mic_tmp3.x = sqrt( pow2(mic3.x - offset_x) + pow2(mic3.y - offset_y) );
    if( mic3.x < offset_x ) { 
      mic_tmp3.x = -mic_tmp3.x;
    } 
    rad = atan( (mic2.y - offset_y) / (mic2.x - offset_x) ); 
  }
  
  poi = calc_mic_arrange_linear_sub( mic_tmp1,mic_tmp2, mic_tmp3, target_z );

  poi1 = rotate( poi, rad );
  poi.y = -poi.y;
  poi2 = rotate( poi, rad );
  poi1.x = offset_x + poi1.x;
  poi1.y = offset_y + poi1.y;
  poi2.x = offset_x + poi2.x;
  poi2.y = offset_y + poi2.y;

  if( test_target_area( poi1 ) ) {
    return poi1;
  } else if( test_target_area( poi2 ) ) {
    return poi2;
  }
  poi1.x = -1;
  poi1.y = -1;
  return poi1;
}

//-----------------------------------------------------------------------------------------
// rotate coordinate
//-----------------------------------------------------------------------------------------
point_of_impact rotate( point_of_impact poi, double rad )
{
  point_of_impact result;

  result.x = cos(rad) * poi.x - sin(rad) * poi.y;
  result.y = sin(rad) * poi.x + cos(rad) * poi.y;
     
  return result;
}

//-----------------------------------------------------------------------------------------
// mic conbination
//-----------------------------------------------------------------------------------------
point_of_impact calc_poi_combination( void )
{
  unsigned char mic_id1,mic_id2,mic_id3;
  point_of_impact poi[12];
  unsigned char poi_index;
  unsigned char number_of_poi;
  point_of_impact avg_poi; 
    
  poi_index = 0;
  for( mic_id1 = 0 ; mic_id1 < NUMBER_OF_MIC ; mic_id1++ ) {
    for( mic_id2 = 1 ; mic_id2 < NUMBER_OF_MIC ; mic_id2++ ) {
      if( mic_id1 >= mic_id2 ) { continue; }  
      for( mic_id3 = 2 ; mic_id3 < NUMBER_OF_MIC ; mic_id3++ ) {
        if( mic_id2 >= mic_id3 ) { continue; }  
        if( test_mic_arrange_liner( mic_id1, mic_id2, mic_id3 ) == true ) {
//        Serial.println( " liner " );
          poi[poi_index] = calc_mic_arrange_linear(     mic[mic_id1], mic[mic_id2], mic[mic_id3], TARGET_Z );
        } else {
          poi[poi_index] = calc_not_mic_arrange_linear( mic[mic_id1], mic[mic_id2], mic[mic_id3], TARGET_Z );
        }
        poi_index++;
      }
    }
  }
  number_of_poi = poi_index;
   
  avg_poi.x = 0.0;
  avg_poi.y = 0.0;
  for( poi_index = 0 ; poi_index < number_of_poi ; poi_index++ ) {
    avg_poi.x = avg_poi.x + poi[poi_index].x;
    avg_poi.y = avg_poi.y + poi[poi_index].y;
  }
  avg_poi.x = avg_poi.x / number_of_poi;
  avg_poi.y = avg_poi.y / number_of_poi;
  #ifdef DEBUG
  for( poi_index = 0 ; poi_index < number_of_poi ; poi_index++ ) {
    dump_circle( poi[poi_index].x, poi[poi_index].y, (poi_index + 1) * 0.2, poi_index + 1 );  // generate NABE CAD script
    Serial.print(" // ");
    dump_double( poi[poi_index].x );
    dump_double( poi[poi_index].y );
    dump_double( poi[poi_index].x - avg_poi.x );
    dump_double( poi[poi_index].y - avg_poi.y );
    Serial.println();
  }
  #endif
  
  return avg_poi;
}

//-----------------------------------------------------------------------------------------
// MAX3485 TX enable
//-----------------------------------------------------------------------------------------
void rs485_TX_enable()
{
  digitalWrite( RS485_CONTROL_PIN, HIGH );
}

//-----------------------------------------------------------------------------------------
// MAX3485 RX enablen
//-----------------------------------------------------------------------------------------
void rs485_RX_enable()
{
  Serial1.flush();
  digitalWrite( RS485_CONTROL_PIN, LOW );
}

//-----------------------------------------------------------------------------------------
// get temperature
//-----------------------------------------------------------------------------------------
double get_temperature()
{
  String str;
  char   buf[10];  
  double temperature;
 
  send_command( COMMAND_QUERY_TEMPERATURE );
  str = Serial1.readStringUntil('\n');
  if( str.length() == 0 ) {
    #ifdef DEBUG
    Serial.println("get_temperature timeout");
    #endif    
    return -300;   // time out
  } else {
    temperature = (double)str.toInt() * 0.0625;
    #ifdef DEBUG
    sprintf( buf, "%4.2f", temperature );
    Serial.print( "temperature " );
    Serial.println( buf );
    #endif
    return temperature;    
  }
}

//-----------------------------------------------------------------------------------------
// send command arduino -> PIC24
//-----------------------------------------------------------------------------------------
void send_command( unsigned char command )
{
  rs485_TX_enable();
  Serial1.write( command );
  rs485_RX_enable();
}


//-----------------------------------------------------------------------------------------
// Distance from the center
//-----------------------------------------------------------------------------------------
point_of_impact distance_from_center( point_of_impact poi )
{
  point_of_impact result;

  result.x = poi.x - TARGET_CENTER_X;
  result.y = poi.y - TARGET_CENTER_Y;

  return result;
}

//-----------------------------------------------------------------------------------------
// setup
//-----------------------------------------------------------------------------------------
void setup() {
  pinMode( PUSH_SW1, INPUT  );
  pinMode( PUSH_SW2, INPUT  );
  pinMode( PUSH_SW3, INPUT  );
  pinMode( DIP_SW1,  INPUT  );
  pinMode( DIP_SW2,  INPUT  );
  pinMode( DIP_SW3,  INPUT  );
  pinMode( DIP_SW4,  INPUT  );
  pinMode( LED1,     OUTPUT );
  pinMode( LED2,     OUTPUT );
  pinMode( RS485_CONTROL_PIN, OUTPUT );

  Serial.begin(38400);   // PC <-> arduino
  Serial1.begin(19200);  // mic <-> arduino
  rs485_RX_enable(); 
  
  Mic_error = false;
  record_no    = 0;
  delay(500);
  
  temperature = get_temperature();
  Serial1.flush();
  Serial.flush();

  tft.init();
  tft.setRotation(3);
  tft.fillScreen(TFT_GREY);

  first_bullet = true;
  draw_AP_target();
}

//-----------------------------------------------------------------------------------------
// loop
//-----------------------------------------------------------------------------------------
void loop() {
  point_of_impact poi;
  point_of_impact draw_poi;
  static point_of_impact prev_poi;
  
  if( Serial1.available()> 0 )  {
    serial_buf = Serial1.readStringUntil('\n');
    if( serial_buf.compareTo("EOR") == 0 ) {   // end of record
      #ifdef DEBUG
      Serial.println( "//----------------------------------------------------" ); 
      #endif
      if( Mic_error == false ) {
        led1_off();
        led2_off();
        for( unsigned char i = 0 ; i < NUMBER_OF_MIC ; i++ ) {
          mic[i].r = calc_distance( timer_count[i], temperature );
          #ifdef DEBUG
          Serial.print( "// " );  
          dump_mm( mic[i].r );
          Serial.print( "mm   " );
          dump_uint( timer_count[i] );
          Serial.println( "count  " );
          #endif
        }
        poi = calc_poi_combination();
        draw_poi = distance_from_center( poi );
        if( !first_bullet ) {
          draw_bullet( prev_poi.x, prev_poi.y, CALIBER, TFT_GREY );
        }
        draw_bullet(  draw_poi.x, draw_poi.y, CALIBER, TFT_RED  );
        prev_poi.x = draw_poi.x;
        prev_poi.y = draw_poi.y;
        first_bullet = false;
        #ifdef DEBUG
        Serial.print("X="); 
        Serial.println(draw_poi.x);
        Serial.print("Y="); 
        Serial.println(draw_poi.y);
        dump_circle( poi.x, poi.y, 3.0, 1 );  // generate NABE CAD script
        Serial.println();
        #endif
      } else {
        #ifdef DEBUG
        led1_on();
        Serial.println("// mic error");
        #endif
      }  
      record_no = 0;
      Mic_error = false;
      send_command( COMMAND_LED1_ON );
      delay(300);
    } else { 
      if( serial_buf.charAt(0) != 'E' ) {  
        timer_count[record_no] = serial_buf.toInt();
        record_no++;
      } else {
        Mic_error = true;
      }
      #ifdef DEBUG      
      Serial.print( "// RAW DATA " );  
      Serial.println( serial_buf );
      #endif       
    }
  } 

  if( digitalRead( PUSH_SW1 ) == SW_OFF ) { switch1_enable = true; }
  if( digitalRead( PUSH_SW2 ) == SW_OFF ) { switch2_enable = true; }
  if( digitalRead( PUSH_SW3 ) == SW_OFF ) { switch3_enable = true; }

  if( sw3_on() ) {
    temperature = get_temperature();
    first_bullet = true;
    draw_AP_target();
  }

  //---- relay sample program ----------------------------------------
  if( sw1_on() ) {
    send_command( COMMAND_RELAY2_ON );
  }

  if( sw2_on() ) {
    send_command( COMMAND_RELAY2_OFF );
  }
  //---- relay sample program E N D ----------------------------------

}
//--------------- E N D -------------------------------------------------------------------
